#include <stdlib.h>
#include <stdio.h>
#include "gauss.h"
#include "moves.h"
#include "poly.h"

int main(int argc, char *argv[]) {
  if (argc < 3) {
    printf("Usage: checkpolys <number of crossings to generate> <number of crossings> [number of strands] [0 for non-cyclic] [1 for reversible]\n");
    return 0;
  }
  int sz = atoi(argv[1]);
  int maxp = atoi(argv[2]);
  int ns = (argc >= 4 ? atoi(argv[3]) : 1);
  cyclic = (argc >= 5 ? atoi(argv[4]) : 1);
  reversible = (argc == 6 ? atoi(argv[5]) : 0);
  initcrossings((2 * ns * ns - ns) * sz);
  initpascal((2 * ns * ns - ns) * sz + 2);
  int n;
  code *codes = genallcodes(maxp, &n);
  //int *lowest = findlowest(codes, n, sz);
  //initlowest(FILENAME, 0);
  //findlowest(codes, n, sz);
  initlowest(sz);
  n <<= sz - maxp;

  int i, m = 0, j = 0;
  for (i = 0; i < n; i++) {
    code c = getcode(codes, sz, i);
    if (c.n != -1) {
      fprintf(stderr, "%d of %d\r", i, n);
      if (c.n <= maxp) {
	j++;
	knot k = getlowest(i);
	if (k == -1) {
	  //printf("*%d* ", i);
	m++;
	} else {
	  //printf("%d ", lowest[i]);
	  code l = getcode(codes, sz, k);
	  poly pc = multijonespoly(c, ns);
	  poly pl = multijonespoly(l, ns);
	  if (comparepoly(pc, pl) != 0) {
	    printf("%d ", i);
	    printpoly(pc);
	    printf(" ");
	    printcode(c);
	    printf("%d ", k);
	    printpoly(pl);
	    printf(" ");
	    printcode(l);
	    printf("\n");
	  }
	  if (Jinvar(c) != Jinvar(l)) {
	    printf("%d %d ", i, Jinvar(c));
	    printcode(c);
	    printf("%d %d ", k, Jinvar(l));
	    printcode(l);
	  }
	  twopoly tpc = genalexander(c);
	  twopoly tpl = genalexander(l);
	  if (comparetwopoly(tpc, tpl) != 0) {
	    printf("%d ", i);
	    printtwopoly(tpc);
	    printf(" ");
	    printcode(c);
	    printf("%d ", k);
	    printtwopoly(tpl);
	    printf(" ");
	    printcode(l);
	    printf("\n");
	  }
	  free(pc.coeffs);
	  free(pl.coeffs);
	  freetwopoly(tpc);
	  freetwopoly(tpl);
	  free(l.code);
	}
      }
      free(c.code);
    }
  }
  printf("Number of codes checked: %d; unique: %d\n", j, m);

  return 0;
}
