#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "gauss.h"
#include "moves.h"
#include "poly.h"

typedef struct intlist { // linked list for storing codes associated with poly
  struct intlist *next;
  int n;
} intlist;

typedef struct polytree { // binary tree for storing polynomials
  struct polytree *left, *right;
  twopoly p;
  intlist *codes;
} polytree;

void addtotree(polytree *, twopoly, int);
int printtree(polytree *, code *, int);

int main(int argc, char *argv[]) {
  if (argc < 3) {
    printf("Usage: counttwopolys <max crossings> <number of crossings for output> [0 for non-cyclic] [1 for reversible]\n");
    return 0;
  }
  int sz = atoi(argv[1]);
  int maxp = atoi(argv[2]);
  //int ns = (argc >= 3 ? atoi(argv[2]) : 1);
  cyclic = (argc >= 4 ? atoi(argv[3]) : 1);
  reversible = (argc == 5 ? atoi(argv[4]) : 0);
  //initcrossings((2 * ns * ns - ns) * sz);
  initcrossings(sz);
  //initpascal((2 * ns * ns - ns) * sz + 2);
  int n;
  code *codes = genallcodes(maxp, &n);
  //int *lowest = findlowest(codes, n, sz);
  //initlowest(FILENAME, 0);
  initlowest(sz);
  n <<= sz - maxp;
  srand(time(0));

  int i, m = 0;
  polytree t;
  t.p.n = -1;
  t.left = t.right = NULL;
  for (i = 0; i < n; i++)
    if (getlowest(i) == -1) {
      fprintf(stderr, "%d of %d\r", i, n);
      code c = getcode(codes, sz, i);
      if (c.n != -1) {
	if (c.n <= maxp) {
	  m++;
	  fflush(stdout);
	  //poly p = multijonespoly(c, ns);
	  twopoly p = genalexander(c);
	  if (t.p.n == -1) { // first one
	    t.p = p;
	    t.codes = calloc(1, sizeof(intlist));
	    t.codes->n = i;
	  } else
	    addtotree(&t, p, i);
	  //free(p.coeffs);
	}
	free(c.code);
      }
    }
  printf("                      \r");
  int j = printtree(&t, codes, sz);
  printf("%d unique codes, with %d different polynomials\n", m, j);
  return 0;
}

void addtotree(polytree *t, twopoly p, int n) {
  int cmp = comparetwopoly(p, t->p);
  if (cmp == 0) {
    intlist *l = t->codes;
    while (l->next != NULL)
      l = l->next;
    l->next = malloc(sizeof(intlist));
    l = l->next;
    l->next = NULL;
    l->n = n;
    freetwopoly(p); // this is only called by main(), and it's appropriate here
    return;
  } else if (cmp == -1) {
    if (t->left != NULL) {
      addtotree(t->left, p, n);
      return;
    } else {
      t->left = calloc(1, sizeof(polytree));
      t = t->left;
    }
  } else {
    if (t->right != NULL) {
      addtotree(t->right, p, n);
      return;
    } else {
      t->right = calloc(1, sizeof(polytree));
      t = t->right;
    }
  }
  t->p = p;
  t->codes = malloc(sizeof(intlist));
  t->codes->next = NULL;
  t->codes->n = n;
}

int printtree(polytree *t, code *codes, int sz) { // returns number of polys
  if (t == NULL)
    return 0;
  int n = 1;
  n += printtree(t->left, codes, sz);
  printtwopoly(t->p);
  printf("\n");
  intlist *l;
  for (l = t->codes; l != NULL; l = l->next) {
    code c = getcode(codes, sz, l->n);
//    printf("%d \t", Jinvar(c));
    printf("\t");
    printcode(c);
    free(c.code);
  }
  n += printtree(t->right, codes, sz);
  return n;
}
