#define _LARGEFILE64_SOURCE
//#define _FILE_OFFSET_BITS 64
#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <limits.h>
#include <fcntl.h>
//#include <sys/mman.h>
#include "gauss.h"
#include "moves.h"

#define FILEPATH "files/"
#define FILENAME FILEPATH "lowest-%d-%d-%d"

int cmin(int, int);
static int lowestfile = -1;

/* knot *getlowest(knot i) { // accesses mmap'd file instead of array */
/*   // maps 1GB at a time (128 million long longs or 256 million unsigned ints) */
/*   static int n = -1, bits = sizeof(knot) == 8 ? 27 : 28; */
/*   if (lowestfile == -1) */
/*     lowestfile = open64(filename, O_RDWR | O_CREAT); */
/*   static knot *lowest = NULL; */
/*   if (i >> bits != n) { */
/*     if (lowest != NULL) */
/*       munmap(lowest, 1 << 30); */
/*     lowest = mmap64(NULL, 1 << 30, */
/* 		    PROT_READ | PROT_WRITE, MAP_SHARED, lowestfile, */
/* 		    (i & ~((off64_t)((1 << bits) - 1))) * sizeof(knot)); */
/*     if (lowest == -1 || lowest == 0) { */
/*       printf("Error %d\n", errno); */
/*       exit(1); */
/*     } */
/*     n = i >> bits; */
/*     /\*if (n > 0) { */
/*       printf("OK\n"); */
/*       exit(0); */
/*       }*\/ */
/*   } */
/*   return lowest + (i & ((1 << bits) - 1)); */
/* } */

knot getlowest(knot i) {
  knot r;
  pread64(lowestfile, &r, sizeof(knot), ((off64_t) i) * sizeof(knot));
  return r;
}

void setlowest(knot i, knot val) {
  pwrite64(lowestfile, &val, sizeof(knot), ((off64_t) i) * sizeof(knot));
}

/* void initlowest(knot n) { */
/*   if (lowestfile == -1) */
/*     lowestfile=open64("/home/jeremy/files/lowest",O_RDWR | O_CREAT | O_TRUNC); */
/*   pwrite64(lowestfile, "", 1, ((off64_t) n) * sizeof(knot) - 1); */
/* } */

void initlowestfile(int sz, int new) {
  char filename[100];
  snprintf(filename, 100, FILENAME, sz, cyclic ? 1 : 0, reversible ? 1 : 0);
  lowestfile = open64(filename, O_RDWR | (new ? (O_CREAT | O_EXCL) : 0), 0644);
  if (lowestfile == -1) {
    perror(NULL);
    exit(1);
  }
}

void initlowest(int sz) {
  initlowestfile(sz, 0);
}

void findlowest(code *codes, knot n, int sz) {
  //int i, *low;
  //low = malloc(n * sizeof(int));
  initlowestfile(sz, 1);
  knot i;
  for (i = 0; i < n; i++) {
    //low[i] = -1;
    //*(getlowest(i)) = -1LL;
    setlowest(i, -1);
    //int equiv[10 * sz], m = 0, j, lowest=INT_MAX;
    int m = 0, j;
    knot equiv[10 * sz], lowest = KNOT_MAX;
    code c = getcode(codes, sz, i);
    if (c.n == -1)
      continue;

    for (j = 0; j < 2 * c.n + (cyclic ? 0 : -1); j++) { // look for R2
      int jj = (j + 1) % (2 * c.n); // deals with cyclic case
      if (crossings[c.code[j]].p == O && crossings[c.code[jj]].p == O) {
	int k;
	if (c.code[j] % 3 - 1 == -(c.code[jj] % 3 - 1)) { // found OxOy
	  for (k = 0; k < 2 * c.n + (cyclic ? 0 : -1); k++) { // find UxUy
	    int kk = (k + 1) % (2 * c.n);
	    if (k != j && k != jj && kk != j &&
		((c.code[k]/6 == c.code[j]/6 && c.code[kk]/6 == c.code[jj]/6)||
		(c.code[k]/6 == c.code[jj]/6 && c.code[kk]/6 == c.code[j]/6))){
	      code r = R2(c, j, k);
	      equiv[m] = find(r, codes, 0, n >> sz);
	      assert(equiv[m] != -1);
	      equiv[m] <<= sz;
	      equiv[m] |= getsigns(r);
	      assert(equiv[m] < i);
	      m++;
	      free(r.code);
	      break;
	    }
	  }
	}

	int n1 = c.code[j]/6, n2=c.code[jj]/6, n3, ju, jju;
	for (k = 0; k < 2 * c.n; k++) { // R3: find Ux,Uy
	  if (k != j && c.code[k]/6 == n1)
	    ju = k;
	  else if (k != jj && c.code[k]/6 == n2)
	    jju = k;
	}

	// ju is at pair of under crossings
	int t, u, m2 = 1, m3 = 1;
	for (k = 0; k < 4; k++) { // 8 variants of R3
	  if (k == 2) {
	    int tmp = j;
	    j = jj;
	    jj = tmp;
	    tmp = ju;
	    ju = jju;
	    jju = tmp;
	    m3 = -1;
	  }
	  if (k % 2 == 0) {
	    t = ju - 1;
	    if (cyclic && t < 0)
	      t = 2 * c.n - 1;
	    m2 = 1;
	  } else {
	    t = ju + 1;
	    if (cyclic && t >= 2 * c.n)
	      t = 0;
	    m2 = -1;
	  }
	  if (t >= 0 && t != jju && t != j && t != jj) {
	    if (crossings[c.code[t]].p == U) {
	      n3 = c.code[t]/6;
	      u = jju - 1; // check for match between Uz, Oz
	      if (cyclic && u < 0)
		u = 2 * c.n - 1;
	      if (u >= 0 && u != t && u != ju && u != j && u != jj) {
		if (c.code[u]/6 == n3) { // found it: OxOy/OzUy/UzUx (x=y=z)
		  if (c.code[j]%3-1 == m3*(c.code[t]%3-1) // signs match
		      && c.code[j]%3-1 == m2*(c.code[jj]%3-1)) {
		    code r = R3(c, cmin(j, jj), cmin(ju, t), u);
		    equiv[m] = find(r, codes, 0, n >> sz);
		    equiv[m] <<= sz;
		    equiv[m] |= getsigns(r);
		    if (equiv[m] < i)
		      m++;
		    free(r.code);
		    continue;
		  }
		}
	      }
	      u = jju + 1;
	      if (cyclic && u >= 2 * c.n)
		u = 0;
	      if (u < 2 * c.n && u != t && u != ju && u != j && u != jj) {
		if (c.code[u]/6 == n3) { // found it: OxOy/UyOz/UzUx (-x=y=z)
		  if (-(c.code[j]%3-1) == m3*(c.code[t]%3-1)
		      && -(c.code[j]%3-1) == m2*(c.code[jj]%3-1)) {
		    code r = R3(c, cmin(j, jj), cmin(ju, t), jju);
		    equiv[m] = find(r, codes, 0, n >> sz);
		    equiv[m] <<= sz;
		    equiv[m] |= getsigns(r);
		    if (equiv[m] < i)
		      m++;
		    free(r.code);
		  }
		}
	      }
	    }
	  }
	}
	int tmp = j;
	j = jj;
	jj = tmp;

      }
    }

    if (m == 0) {
      code r = first(c);
      if (r.code != c.code) {
	equiv[0] = find(r, codes, 0, n >> sz);
	equiv[0] <<= sz;
	equiv[0] |= getsigns(r);
	if (equiv[0] < i)
	  m++;
	free(r.code);
      }
    }

    for (j = 0; j < m; j++) { // trace down to base codes
      knot l = equiv[j];
      knot k;
      do {
	k = l;
	//l = low[k];
	//l = *(getlowest(k));
	l = getlowest(k);
      } while (l != -1);
      equiv[j] = k;
      if (k < lowest)
	lowest = k;
    }

    for (j = 0; j < m; j++)
      if (equiv[j] > lowest)
	//low[equiv[j]] = lowest;
	//*(getlowest(equiv[j])) = lowest;
	setlowest(equiv[j], lowest);
    if (lowest != KNOT_MAX) //INT_MAX)
      //low[i] = lowest;
      //*(getlowest(i)) = lowest;
      setlowest(i, lowest);

    free(c.code);
  }

  for (i = 0; i < n; i++) { // make sure everything points to the base code
    knot j = getlowest(i);//low[i];
    if (j != -1) {
      knot k = getlowest(j);//low[j];
      if (k != -1) {
	/*printf("%d ", i);
	code c = getcode(codes, sz, i);
	printcode(c);
	free(c.code);
	printf("%d ", j);
	c = getcode(codes, sz, j);
	printcode(c);
	free(c.code);
	printf("%d ", k);
	c = getcode(codes, sz, k);
	printcode(c);
	free(c.code);
	printf("\n");*/

	//low[i] = k;
	//*(getlowest(i)) = k;
	setlowest(i, k);
      }
    }
  }

  //return low;
}

int cmin(int a, int b) { // "cyclic" min; helper for findlowest
  if (a == b + 1)
    return b;
  if (b == a + 1)
    return a;
  if (a > b)
    return a;
  return b;
}

knot getknot(code c, code *codes, int n, int sz) {
  knot k = find(c, codes, 0, n >> sz);
  k <<= sz;
  k |= getsigns(c);
  if (getlowest(k) != -1)
    k = getlowest(k);
  return k;
}

code R2(code c, int a, int b) { // perform R2 move on c, at indices a, b
  if (a > b) {
    int tmp = b;
    b = a;
    a = tmp;
  }
  code r;
  r.n = c.n - 2;
  r.code = malloc(2 * r.n * sizeof(codeelem));
  int i, n = 0;
  for (i = (b == 2 * c.n - 1 ? 1 : 0); i < a; i++)
    r.code[n++] = c.code[i];
  for (i = a + 2; i < b; i++)
    r.code[n++] = c.code[i];
  for (i = b + 2; i < 2 * c.n; i++)
    r.code[n++] = c.code[i];

  code s = simplify(r);
  if (s.code != r.code)
    free(r.code);
  return s;
}

code R3(code c, int x, int y, int z) { // perform R3 on c, at indices x, y, z
  code r;
  r.n = c.n;
  r.code = malloc(2 * r.n * sizeof(codeelem));
  memcpy(r.code, c.code, 2 * r.n * sizeof(codeelem));

  int tmp, n = 2 * r.n;
  tmp = r.code[x];
  r.code[x] = r.code[(x + 1) % n];
  r.code[(x + 1) % n] = tmp;

  tmp = r.code[y];
  r.code[y] = r.code[(y + 1) % n];
  r.code[(y + 1) % n] = tmp;

  tmp = r.code[z];
  r.code[z] = r.code[(z + 1) % n];
  r.code[(z + 1) % n] = tmp;

  code s = simplify(r);
  if (s.code != r.code)
    free(r.code);
  return s;
}


code inverse(code c) {
  code a;
  a.n = c.n;
  a.code = malloc(2 * a.n * sizeof(codeelem));

  int i;
  for (i = 0; i < 2 * a.n; i++)
    a.code[i] = c.code[2 * a.n - 1 - i];
  fixnums(a, a.n);
  int tmp = reversible;
  reversible = 0;
  code r = first(a);
  reversible = tmp;
  if (r.code != a.code)
    free(a.code);
  return r;
}

code vmirror(code c) { // vertical "switching crossings" mirror
  code a;
  a.n = c.n;
  a.code = malloc(2 * a.n * sizeof(codeelem));

  int i;
  for (i = 0; i < 2 * a.n; i++) {
    if (crossings[c.code[i]].p == O)
      a.code[i] = c.code[i] + 3 * (U - O);
    else
      a.code[i] = c.code[i] + 3 * (O - U);
    if (a.code[i] % 3 == 2)
      a.code[i] -= 2;
    else
      a.code[i] += 2;
  }
  code r = first(a);
  if (r.code != a.code)
    free(a.code);
  return r;
}

code hmirror(code c) { // horizontal "switching signs" mirror
  code a;
  a.n = c.n;
  a.code = malloc(2 * a.n * sizeof(codeelem));

  int i;
  for (i = 0; i < 2 * a.n; i++)
    if (c.code[i] % 3 == 0)
      a.code[i] = c.code[i] + 2;
    else
      a.code[i] = c.code[i] - 2;
  code r = first(a);
  if (r.code != a.code)
    free(a.code);
  return r;
}

code doublemirror(code c) { // vmirror(hmirror(c)); amounts to O <--> U
  code a;
  a.n = c.n;
  a.code = malloc(2 * a.n * sizeof(codeelem));

  int i;
  for (i = 0; i < 2 * a.n; i++)
    if (crossings[c.code[i]].p == O)
      a.code[i] = c.code[i] + 3 * (U - O);
    else
      a.code[i] = c.code[i] + 3 * (O - U);
  code r = first(a);
  if (r.code != a.code)
    free(a.code);
  return r;
}

// returns -1 if this isn't the first
// otherwise, return value has x & 4 true if the knot has a distinct inverse
// and the least significant two bits indicate the number of mirrors
int isfirst(knot i, code *codes, int n, int sz) {
  code c = getcode(codes, sz, i);

  code a, b;
  int inv = 0, mirrors;
  knot k, d;
  
  if (!reversible) { // inverse
    a = inverse(c);
    k = getknot(a, codes, n, sz);
    if (k < i) {
      free(a.code);
      free(c.code);
      return -1;
    }
    if (k == i)
      inv = 0;
    else
      inv = 1;
    b = vmirror(a);
    k = getknot(b, codes, n, sz);
    free(b.code);
    if (k < i) {
      free(a.code);
      free(c.code);
      return -1;
    }
    b = hmirror(a);
    k = getknot(b, codes, n, sz);
    free(b.code);
    if (k < i) {
      free(a.code);
      free(c.code);
      return -1;
    }
    b = doublemirror(a);
    k = getknot(b, codes, n, sz);
    free(b.code);
    if (k < i) {
      free(a.code);
      free(c.code);
      return -1;
    }
    free(a.code);
  }
  
  // mirrors
  a = vmirror(c);
  k = getknot(a, codes, n, sz);
  free(a.code);
  if (k < i) {
    free(c.code);
    return -1;
  }
  if (k == i)
    mirrors = 0;
  else
    mirrors = 1;
  
  a = hmirror(c);
  d = getknot(a, codes, n, sz);
  free(a.code);
  if (d < i) {
    free(c.code);
    return -1;
  }
  if (mirrors == 0)
    if (d == i)
      mirrors = 0;
    else
      mirrors = 2;
  else
    if (d != i && d != k) {
      mirrors = 7;
      
      a = doublemirror(c);
      k = getknot(a, codes, n, sz);
      free(a.code);
      if (k < i) {
	free(c.code);
	return -1;
      }
    } else if (d != i)
      mirrors = 3;
    else
      mirrors = 1;

  return (inv << 3) | mirrors;
}

// returns first of the mirrors/inversions
knot getfirst(knot i, code *codes, int n, int sz) {
  knot f;
  if (getlowest(i) != -1)
    f = getlowest(i);
  else
    f = i;
  code c = getcode(codes, sz, i);

  code a, b;
  knot k;
  
  if (!reversible) { // inverse
    a = inverse(c);
    k = getknot(a, codes, n, sz);
    if (k < f)
      f = k;

    b = vmirror(a);
    k = getknot(b, codes, n, sz);
    free(b.code);
    if (k < f)
      f = k;

    b = hmirror(a);
    k = getknot(b, codes, n, sz);
    free(b.code);
    if (k < f)
      f = k;

    b = doublemirror(a);
    k = getknot(b, codes, n, sz);
    free(b.code);
    if (k < f)
      f = k;
    free(a.code);
  }
  
  // mirrors
  a = vmirror(c);
  k = getknot(a, codes, n, sz);
  free(a.code);
  if (k < f)
    f = k;
  
  a = hmirror(c);
  k = getknot(a, codes, n, sz);
  free(a.code);
  if (k < f)
    f = k;

  a = doublemirror(c);
  k = getknot(a, codes, n, sz);
  free(a.code);
  if (k < i)
    f = k;

  free(c.code);
  return f;
}
