#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include "gauss.h"
#include "poly.h"
#include "pd.h"

typedef struct link {
  code *codes;  // codes[i].n will indicate the number of codeelems in that
  int n;                // component, instead of the usual number of crossings
} link;

twopoly det(twopoly ***, int, twopoly *);
void detr(twopoly ***, int, twopoly *, int *, int, int *, int **);
poly addpoly(poly, poly);
poly multpoly(poly, poly);
twopoly addtwopoly(twopoly, twopoly);
twopoly multtwopoly(twopoly, twopoly);
poly loopspoly(int);
void cprintpoly(poly, char);
char *scprintpoly(poly, char);
poly bracket(link, int);
poly bracket2(PD);
PD link2PD(link);
int writhe(code);
int **pascaltriangle;
int sortweb(unsigned char **, int *, poly *, int, int, int);

poly jonespoly(code c) { // (-A^3)^(-writhe(c)) * bracket(c)
  link l;
  l.n = 1;
  l.codes = &c;
  c.n *= 2;
  poly p = bracket(l, 0);
  c.n /= 2;
  int w = writhe(c);
  p.m += -3 * w;
  if (w % 2 != 0) { // multiply p by -1
    int i;
    for (i = 0; i < p.n; i++)
      p.coeffs[i] *= -1;
  }
  return p;
}

poly jonespoly2(code c) {
  int w = writhe(c);
  PD pd = code2PD(c);
  poly p = bracket2(pd);
  freePD(pd);
  p.m += -3 * w;
  if (w % 2 != 0) { // multiply p by -1
    int i;
    for (i = 0; i < p.n; i++)
      p.coeffs[i] *= -1;
  }
  return p;
}

poly multijonespoly(code c, int n) { // jones polynomial of multi stranded knot
  if (c.n == 0)
    return loopspoly(n - 1);
  int w = writhe(c);
  //int ln = (n * n - n) * w; // linking number
  link l;
  l.n = n;
  l.codes = malloc(n * sizeof(code));
  /*l.codes[0].n = l.codes[1].n = c.n * 4 + abs(ln);
  l.codes[0].code = malloc(l.codes[0].n * sizeof(codeelem));
  l.codes[1].code = malloc(l.codes[0].n * sizeof(codeelem));*/
  int i, j, k, a;
  for (i = 0; i < n; i++) {
    l.codes[i].n = c.n * 2 * n + abs(2 * (n - 1) * w);
    l.codes[i].code = malloc(l.codes[i].n * sizeof(codeelem));
  }
  for (i = 0; i < 2 * c.n; i++) {
    // work as if this is a positive crossing, then swap if it's negative
    int sp = c.code[i] % 6, nn = c.code[i] / 6;
    if (sp / 3 == O) { // over crossing
      /*l.codes[0].code[2 * i] = 24 * n + sp;
      l.codes[0].code[2 * i + 1] = 24 * n + 6 + sp;
      l.codes[1].code[2 * i] = 24 * n + 12 + sp;
      l.codes[1].code[2 * i + 1] = 24 * n + 18 + sp;*/
      if (sp % 3 == 2) // positive crossing
	for (j = 0, a = 0; j < n; j++)
	  for (k = 0; k < n; k++, a++)
	    l.codes[j].code[n * i + k] = (n * n * nn + a) * 6 + sp;
      else
	for (j = 0, a = 0; j < n; j++)
	  for (k = n - 1; k >= 0; k--, a++)
	    l.codes[j].code[n * i + k] = (n * n * nn + a) * 6 + sp;
    } else { // under crossing
      /*l.codes[0].code[2 * i] = 24 * n + 12 + sp;
      l.codes[0].code[2 * i + 1] = 24 * n + sp;
      l.codes[1].code[2 * i] = 24 * n + 18 + sp;
      l.codes[1].code[2 * i + 1] = 24 * n + 6 + sp;*/
      if (sp % 3 == 2) // positive crossing
	for (k = n - 1, a = 0; k >= 0; k--)
	  for (j = 0; j < n; j++, a++)
	    l.codes[j].code[n * i + k] = (n * n * nn + a) * 6 + sp;
      else
	for (k = 0, a = 0; k < n; k++)
	  for (j = 0; j < n; j++, a++)
	    l.codes[j].code[n * i + k] = (n * n * nn + a) * 6 + sp;
    }
    /*if (sp % 3 == 0) { // negative crossing
      int tmp = l.codes[0].code[2 * i];
      l.codes[0].code[2 * i] = l.codes[0].code[2 * i + 1];
      l.codes[0].code[2 * i + 1] = tmp;
      tmp = l.codes[1].code[2 * i];
      l.codes[1].code[2 * i] = l.codes[1].code[2 * i + 1];
      l.codes[1].code[2 * i + 1] = tmp;
      }*/
  }
  if (w != 0) { // make the linking number 0
    if (w < 0)
      i = 2;
    else
      i = 0;
    /*int n;
    for (n = 4 * c.n; n < l.codes[0].n; n+=2) {
      l.codes[0].code[n] = 6 * n + 3 * (i == 0 ? U : O) + i;
      l.codes[0].code[n + 1] = 6 * (n + 1) + 3 * (i == 0 ? O : U) + i;
      l.codes[1].code[n] = 6 * n + 3 * (i == 0 ? O : U) + i;
      l.codes[1].code[n + 1] = 6 * (n + 1) + 3 * (i == 0 ? U : O) + i;
      }*/
    int nn = c.n * n * n;
    for (a = 0; a < abs(w); a++)
      for (j = 0; j < n; j++)
	for (k = j + 1; k < n; k++) {
	  l.codes[j].code[2 * c.n * n + 2 * (n - 1) * a +  k - 1] =
	    6 * nn + 3 * (i == 0 ? U : O) + i;
	  l.codes[j].code[2 * c.n * n + 2 * (n - 1) * (a + 1) - k] =
	    6 * (nn + 1) + 3 * (i == 0 ? O : U) + i;
	  l.codes[k].code[2 * c.n * n + 2 * (n - 1) * a + j] =
	    6 * nn + 3 * (i == 0 ? O : U) + i;
	  l.codes[k].code[2 * c.n * n + 2 * (n - 1) * (a + 1) - j - 1] =
	    6 * (nn + 1) + 3 * (i == 0 ? U : O) + i;
	  nn += 2;
	}
  }
  //poly p = bracket(l, 0);
  PD pd = link2PD(l);
  poly p = bracket2(pd);
  freePD(pd);
  p.m += -3 * n * w; // writhe multiplied by n... because it works

  if ((n * w) % 2 != 0) { // multiply p by -1
    for (i = 0; i < p.n; i++)
      p.coeffs[i] *= -1;
  }
  /*free(l.codes[0].code);
    free(l.codes[1].code);*/
  for (i = 0; i < l.n; i++)
    free(l.codes[i].code);
  free(l.codes);
  return p;
}

PD link2PD(link l) {
  int i, j, k;
  PD p;
  p.n = 0;
  for (i = 0; i < l.n; i++)
    p.n += l.codes[i].n;
  p.n /= 2;
  p.crossings = malloc(p.n * sizeof(int *));
  for (i = 0; i < p.n; i++)
    p.crossings[i] = malloc(4 * sizeof(int));
  for (i = 0, k = 0; i < l.n; i++)
    for(j = 0; j < l.codes[i].n; j++, k++) {
      int jj = (j + 1) % l.codes[i].n;
      int n = l.codes[i].code[j] / 6, nn = l.codes[i].code[jj] / 6;
      if (l.codes[i].code[j] % 3 == 0) // outgoing from negative crossing
	if (crossings[l.codes[i].code[j]].p == O)
	  p.crossings[n][3] = k;
	else
	  p.crossings[n][2] = k;
      else // outgoing from positive crossings
	if (crossings[l.codes[i].code[j]].p == O)
	  p.crossings[n][1] = k;
	else
	  p.crossings[n][2] = k;
      if (l.codes[i].code[jj] % 3 == 0) // incoming to negative crossing
	if (crossings[l.codes[i].code[jj]].p == O)
	  p.crossings[nn][1] = k;
	else
	  p.crossings[nn][0] = k;
      else // incoming to positive crossing
	if (crossings[l.codes[i].code[jj]].p == O)
	  p.crossings[nn][3] = k;
	else
	  p.crossings[nn][0] = k;
    }
  return p;
}

int webcompare(const void *a1, const void *a2) {
  //int *a = *((int **) a1), *b = *((int **) a2);
  unsigned char *a = (unsigned char *) a1, *b = (unsigned char *) a2;
  if (a[0] < b[0])
    return -1;
  if (a[0] > b[0])
    return 1;
  if (a[1] < b[1])
    return -1;
  if (a[1] > b[1])
    return 1;
  return 0;
}

void printweb(int ***web, int *webns, poly *polys, int nweb) { //FIXME
  int i, j;
  for (i = 0; i < nweb; i++) {
    printpoly(polys[i]);
    for (j = 0; j < webns[i]; j++)
      printf(" P(%d,%d)", web[i][2 * j], web[i][2 * j + 1]);
    printf("\n");
  }
}

poly bracket2(PD p) { 
  int nweb = 1, nfront = 0;
  poly *polys = malloc(sizeof(poly));
  polys[0].n = 1;
  polys[0].m = 0;
  polys[0].coeffs = malloc(sizeof(int));
  polys[0].coeffs[0] = 1;
  unsigned char **web = malloc(sizeof(unsigned char *));
  int *webns = calloc(1, sizeof(int));
  int i, j, k, l;
  web[0] = malloc(p.n * 4 * sizeof(unsigned char));
  //for (i = 0; i < p.n * 2; i++)
  //web[0][i] = malloc(2 * sizeof(int));
  int *front = malloc(2 * p.n * sizeof(int));
  int *used = calloc(p.n, sizeof(int));
  int nused;
  for (nused = 0; nused < p.n; nused++) {
    // choose crossing with most edges in front
    int most = -1, best; // nbest = 0, bests[p.n];
    for (i = 0; i < p.n; i++)
      if (!used[i]) {
	int e = 0;
	for (j = 0; j < nfront; j++)
	  for (k = 0; k < 4; k++)
	    if (p.crossings[i][k] == front[j]) {
	      e++;
	      break;
	    }
	if (e > most) {
	  most = e;
	  best = i;
	  //bests[0] = i;
	  //nbest = 1;
	} //else if (e == most)
	  //bests[nbest++] = i;
	if (most == 4)
	  break;
      }
    //best = bests[rand() % nbest];
    used[best] = 1;
    for (i = 0; i < 4; i++) { // add the edges in best to the front
      int found = 0;
      for (j = 0; j < nfront; j++)
	if (p.crossings[best][i] == front[j]) {
	  found = 1;
	  break;
	}
      if (!found)
	front[nfront++] = p.crossings[best][i];
    }
    // make space
    web = realloc(web, 2 * nweb * sizeof(unsigned char *));
    for (i = nweb; i < 2 * nweb; i++) {
      web[i] = malloc(p.n * 4 * sizeof(unsigned char));
      //for (j = 0; j < p.n * 2; j++)
      //web[i][j] = malloc(2 * sizeof(int));
    }
    webns = realloc(webns, 2 * nweb * sizeof(int));
    polys = realloc(polys, 2 * nweb * sizeof(poly));
    for (i = 0; i < nweb; i++) { // duplicate the web
      for (j = 0; j < 2 * webns[i]; j++) {
	web[i + nweb][j] = web[i][j];
	//web[i + nweb][j][0] = web[i][j][0];
	//web[i + nweb][j][1] = web[i][j][1];
      }
      polys[i + nweb].n = polys[i].n;
      polys[i + nweb].m = polys[i].m;
      polys[i + nweb].coeffs = malloc(polys[i].n * sizeof(int));
      memcpy(polys[i + nweb].coeffs, polys[i].coeffs, polys[i].n*sizeof(int));
    }
    memcpy(webns + nweb, webns, nweb * sizeof(int));
    for (i = 0; i < nweb; i++) { // for convenience, make web[][][0]<web[][][1]
      if (p.crossings[best][0] < p.crossings[best][1]) {
	web[i][2 * webns[i]] = p.crossings[best][0];
	web[i][2 * webns[i]++ + 1] = p.crossings[best][1];
      } else {
	web[i][2 * webns[i]] = p.crossings[best][1];
	web[i][2 * webns[i]++ + 1] = p.crossings[best][0];
      }
      if (p.crossings[best][2] < p.crossings[best][3]) {
	web[i][2 * webns[i]] = p.crossings[best][2];
	web[i][2 * webns[i]++ + 1] = p.crossings[best][3];
      } else {
	web[i][2 * webns[i]] = p.crossings[best][3];
	web[i][2 * webns[i]++ + 1] = p.crossings[best][2];
      }
      polys[i].m++;
    }
    for (i = nweb; i < 2 * nweb; i++) {
      if (p.crossings[best][0] < p.crossings[best][3]) {
	web[i][2 * webns[i]] = p.crossings[best][0];
	web[i][2 * webns[i]++ + 1] = p.crossings[best][3];
      } else {
	web[i][2 * webns[i]] = p.crossings[best][3];
	web[i][2 * webns[i]++ + 1] = p.crossings[best][0];
      }
      if (p.crossings[best][1] < p.crossings[best][2]) {
	web[i][2 * webns[i]] = p.crossings[best][1];
	web[i][2 * webns[i]++ + 1] = p.crossings[best][2];
      } else {
	web[i][2 * webns[i]] = p.crossings[best][2];
	web[i][2 * webns[i]++ + 1] = p.crossings[best][1];
      } 
      polys[i].m--;
    }
    nweb *= 2;
    for (i = 0; i < nweb; i++) {
      int changed;
      do { // P(a, b) P(b, c) ==> P(a,c)
	// populate array rev[j] = k such tha web[i][k][] == j; and
	//   if a collision is found, then combine as necessary 
	changed = 0;
	int rev[p.n * 2];
	for (j = 0; j < 2 * p.n; j++)
	  rev[j] = -1;
	for (j = 0; j < webns[i]; j++) {
	  int same = -1;
	  for (k = 0; k < 2; k++) {
	    if (rev[web[i][2 * j + k]] == -1)
	      rev[web[i][2 * j + k]] = j;
	    else {
	      if (j != rev[web[i][2 * j + k]])
		same = rev[web[i][2 * j + k]];
	      break;
	    }
	  }
	  if (same != -1) {
	    if (web[i][2 * same] == web[i][2 * j + k])
	      web[i][2 * same] = web[i][2 * j + (k + 1) % 2];
	    else
	      web[i][2 * same + 1] = web[i][2 * j + (k + 1) % 2];
	    rev[web[i][2 * j + (k + 1) % 2]] = same;
	    if (web[i][2 * same] > web[i][2 * same + 1]) {
	      int tmp = web[i][2 * same];
	      web[i][2 * same] = web[i][2 * same + 1];
	      web[i][2 * same + 1] = tmp;
	    }
	    for (l = 2 * j; l < 2 * (webns[i] - 1); l++) {
	      //web[i][l][0] = web[i][l + 1][0];
	      //web[i][l][1] = web[i][l + 1][1];
	      web[i][l] = web[i][l + 2];
	    }
	    webns[i]--;
	    j--;
	    changed = 1;
	  }
	}
      } while (changed);
      int loops = 0;
      for (j = 0; j < webns[i]; j++) // P(a,a) ==> -A^2-A^-2
	if (web[i][2 * j] == web[i][2 * j + 1]) {
	  loops++;
	  webns[i]--;
	  for (k = 2 * j; k < 2 * webns[i]; k++) {
	    //web[i][k][0] = web[i][k + 1][0];
	    //web[i][k][1] = web[i][k + 1][1];
	    web[i][k] = web[i][k + 2];
	  }
	  j--;
	}
      if (webns[i] == 0)
	loops--;
      if (loops > 0) {
	poly lp = loopspoly(loops);
	poly p = multpoly(lp, polys[i]);
	free(lp.coeffs);
	free(polys[i].coeffs);
	polys[i] = p;
      }
      // sort web[i]
      qsort(web[i], webns[i], 2 * sizeof(unsigned char), webcompare);
    }
    //collect like terms, etc.
    nweb = sortweb(web, webns, polys, 0, nweb, p.n * 2);
    web = realloc(web, nweb * sizeof(unsigned char **));
    webns = realloc(webns, nweb * sizeof(int));
    polys = realloc(polys, nweb * sizeof(poly));
  }
  assert(nweb == 1);
  assert(webns[0] == 0);
  poly ret = polys[0];
  //int min = 1 << 20, max = - (1 << 20);
  free(webns);
  free(polys);
  for (i = 0; i < nweb; i++) {
    //for (j = 0; j < 2 * p.n; j++)
    //free(web[i][j]);
    free(web[i]);
  }
  free(web);
  free(used);
  free(front);

  return ret;
}

// mergesort plus collects like terms
int sortweb(unsigned char **web, int *webns, poly *polys,int a,int b,int pn) {
  if (b - a <= 1)
    return b - a;
  int m = (a + b) / 2, xn, yn;
  xn = sortweb(web, webns, polys, a, m, pn);
  yn = sortweb(web, webns, polys, m, b, pn);
  // merge
  unsigned char **webtmp = malloc((b - a) * sizeof(unsigned char *));
  memcpy(webtmp, web + a, (b - a) * sizeof(unsigned char *));
  int *webnstmp = malloc((b - a) * sizeof(int));
  memcpy(webnstmp, webns + a, (b - a) * sizeof(int));
  poly *polystmp = malloc((b - a) * sizeof(poly));
  memcpy(polystmp, polys + a, (b - a) * sizeof(poly));
  int x = 0, y = m - a, z = a, i;
  while (x < xn && y < yn + m - a) {
    int cmp = 0;
    if (webnstmp[x] < webnstmp[y])
      cmp = -1;
    else if (webnstmp[x] > webnstmp[y])
      cmp = 1;
    else
      for (i = 0; i < 2 * webnstmp[x]; i++) {
	if (webtmp[x][i] < webtmp[y][i]) {
	  cmp = -1;
	  break;
	} if (webtmp[x][i] > webtmp[y][i]) {
	  cmp = 1;
	  break;
	}
	/*if (webtmp[x][i][0] < webtmp[y][i][0]) {
	  cmp = -1;
	  break;
	} if (webtmp[x][i][0] > webtmp[y][i][0]) {
	  cmp = 1;
	  break;
	} if (webtmp[x][i][1] < webtmp[y][i][1]) {
	  cmp = -1;
	  break;
	} if (webtmp[x][i][1] > webtmp[y][i][1]) {
	  cmp = 1;
	  break;
	  }*/
      }
    if (cmp < 0) {
      web[z] = webtmp[x];
      webns[z] = webnstmp[x];
      polys[z] = polystmp[x];
      x++;
    } else if (cmp > 0) {
      web[z] = webtmp[y];
      webns[z] = webnstmp[y];
      polys[z] = polystmp[y];
      y++;
    } else {
      web[z] = webtmp[x];
      //for (i = 0; i < pn; i++)
      //free(webtmp[y][i]);
      free(webtmp[y]);
      webns[z] = webnstmp[x];
      polys[z] = addpoly(polystmp[x], polystmp[y]);
      free(polystmp[x].coeffs);
      free(polystmp[y].coeffs);
      x++; y++;
    }
    z++;
  }
  for (; x < xn; x++, z++) {
    web[z] = webtmp[x];
    webns[z] = webnstmp[x];
    polys[z] = polystmp[x];
  }
  for (; y < yn + m - a; y++, z++) {
    web[z] = webtmp[y];
    webns[z] = webnstmp[y];
    polys[z] = polystmp[y];
  }

  free(webtmp);
  free(webnstmp);
  free(polystmp);
  return z - a;
}

poly bracket(link l, int n) {
  int a1 = -1, b1, a2, b2, o, i, j, maxn = 0;

  for (i = 0; i < l.n; i++) {
    for (j = 0; j < l.codes[i].n; j++) {
      if (l.codes[i].code[j] / 6 > maxn)
	maxn = l.codes[i].code[j] / 6;
      if (l.codes[i].code[j] / 6 == n) {
	if (a1 == -1) {
	  a1 = i;
	  b1 = j;
	} else {
	  a2 = i;
	  b2 = j;
	}
      }
    }
  }
  if (a1 == -1) // no more smoothings
    return loopspoly(l.n - 1);

  o = l.codes[a1].code[b1] % 3 - 1;
  poly p1, p2;
  if (a1 == a2) { // same component
    link k; // "with orientation" smoothing;
    k.n = l.n + 1;
    k.codes = malloc(k.n * sizeof(code));
    for (i = 0; i < l.n; i++)
      if (i != a1) {
	k.codes[i].n = l.codes[i].n;
	k.codes[i].code = malloc(k.codes[i].n * sizeof(codeelem));
	memcpy(k.codes[i].code, l.codes[i].code,k.codes[i].n*sizeof(codeelem));
      }
    if (b1 > b2) {
      int tmp = b1;
      b1 = b2;
      b2 = tmp;
    }
    k.codes[a1].n = b2 - b1 - 1;
    k.codes[a1].code = malloc(k.codes[a1].n * sizeof(codeelem));
    for (i = b1 + 1; i < b2; i++)
      k.codes[a1].code[i - b1 - 1] = l.codes[a1].code[i];
    k.codes[l.n].n = l.codes[a1].n - k.codes[a1].n - 2;
    k.codes[l.n].code = malloc(k.codes[l.n].n * sizeof(codeelem));
    for (i = b2 + 1, j = 0; i < l.codes[a1].n; i++, j++)
      k.codes[l.n].code[j] = l.codes[a1].code[i];
    for (i = 0; i < b1; i++, j++) // must be cyclic
      k.codes[l.n].code[j] = l.codes[a1].code[i];
    p1 = bracket(k, n + 1);
    p1.m += o;
    free(k.codes[a1].code);
    free(k.codes[l.n].code);

    // "against orientation" smoothing;
    k.n = l.n;
    k.codes[a1].n = l.codes[a1].n - 2;
    k.codes[a1].code = malloc(k.codes[a1].n * sizeof(codeelem));
    for (i = b1 + 1, j = 0; i < b2; i++, j++)
      k.codes[a1].code[j] = l.codes[a1].code[i];
    int rev = j; //needed later
    for (i = b1 - 1; i >= 0; i--, j++)
      k.codes[a1].code[j] = l.codes[a1].code[i];
    for (i = l.codes[a1].n - 1; i > b2; i--, j++)
      k.codes[a1].code[j] = l.codes[a1].code[i];
    // this messes around with signs, so they need to be fixed
    int signs[maxn + 1];
    for (i = 0; i <= maxn; i++)
      signs[i] = 0;
    for (i = rev; i < k.codes[a1].n; i++)
      signs[k.codes[a1].code[i] / 6] = k.codes[a1].code[i] % 3 - 1;
    for (i = rev; i < k.codes[a1].n; i++)
      signs[k.codes[a1].code[i] / 6] *= -1;
    for (i = 0; i < k.n; i++)
      for (j = 0; j < k.codes[i].n; j++)
	if (signs[k.codes[i].code[j] / 6] != 0)
	  k.codes[i].code[j] = (k.codes[i].code[j] / 3) * 3
	    + signs[k.codes[i].code[j] / 6] + 1;
    p2 = bracket(k, n + 1);
    p2.m -= o;
    for (i = 0; i < l.n; i++)
      free(k.codes[i].code);
    free(k.codes);

  } else { // combining two separate components
    link k;
    k.n = l.n - 1;
    k.codes = malloc(k.n * sizeof(code));
    for (i = 0, j = 0; i < l.n; i++)
      if (i != a1 && i != a2) {
	k.codes[j].n = l.codes[i].n;
	k.codes[j].code = malloc(k.codes[j].n * sizeof(codeelem));
	memcpy(k.codes[j].code, l.codes[i].code,k.codes[j].n*sizeof(codeelem));
	j++;
      }
    k.codes[k.n - 1].n = l.codes[a1].n + l.codes[a2].n - 2;
    k.codes[k.n - 1].code = malloc(k.codes[k.n - 1].n * sizeof(codeelem));

    // "with orientation" smoothing
    for (i = b1 + 1, j = 0; i < l.codes[a1].n; i++, j++)
      k.codes[k.n - 1].code[j] = l.codes[a1].code[i];
    for (i = 0; i < b1; i++, j++)
      k.codes[k.n - 1].code[j] = l.codes[a1].code[i];
    int rev = j; // used later
    for (i = b2 + 1; i < l.codes[a2].n; i++, j++)
      k.codes[k.n - 1].code[j] = l.codes[a2].code[i];
    for (i = 0; i < b2; i++, j++)
      k.codes[k.n - 1].code[j] = l.codes[a2].code[i];
    p1 = bracket(k, n + 1);
    p1.m += o;

    // "against orientation" smoothing
    for (i = b2 - 1, j = rev; i >= 0; i--, j++)
      k.codes[k.n - 1].code[j] = l.codes[a2].code[i];
    for (i = l.codes[a2].n - 1; i > b2; i--, j++)
      k.codes[k.n - 1].code[j] = l.codes[a2].code[i];
    // need to fix signs
    int signs[maxn + 1];
    for (i = 0; i <= maxn; i++)
      signs[i] = 0;
    for (i = rev; i < k.codes[k.n - 1].n; i++)
      signs[k.codes[k.n - 1].code[i] / 6] = k.codes[k.n - 1].code[i] % 3 - 1;
    for (i = rev; i < k.codes[k.n - 1].n; i++)
      signs[k.codes[k.n - 1].code[i] / 6] *= -1;
    for (i = 0; i < k.n; i++)
      for (j = 0; j < k.codes[i].n; j++)
	if (signs[k.codes[i].code[j] / 6] != 0)
	  k.codes[i].code[j] = (k.codes[i].code[j] / 3) * 3
	    + signs[k.codes[i].code[j] / 6] + 1;
    p2 = bracket(k, n + 1);
    p2.m -= o;

    for (i = 0; i < k.n; i++)
      free(k.codes[i].code);
    free(k.codes);
  }
  poly p = addpoly(p1, p2);
  free(p1.coeffs);
  free(p2.coeffs);
  return p;
}

int writhe(code c) {
  int signs[c.n], i, ret = 0;
  for (i = 0; i < 2 * c.n; i++)
    signs[c.code[i] / 6] = c.code[i] % 3 - 1;
  for (i = 0; i < c.n; i++)
    ret += signs[i];
  return ret;
}

poly jonesmirror(poly p) { // A -> A^-1
  poly r;
  r.n = p.n;
  r.m = -(p.n - 1 + p.m);
  r.coeffs = malloc(r.n * sizeof(int));
  int i;
  for (i = 0; i < r.n; i++)
    r.coeffs[i] = p.coeffs[p.n - 1 - i];
  return r;
}

int Jinvar(code c) {
  int signs[c.n], dists[c.n], i, j, ret = 0;
  for (i = 0; i < 2 * c.n; i++) {
    signs[c.code[i] / 6] = c.code[i] % 3 - 1;
    for (j = i + 1; j < 2 * c.n; j++)
      if (c.code[i] / 6 == c.code[j] / 6) {
	dists[c.code[i] / 6] = j - i;
	break;
      }
  }
  for (i = 0; i < c.n; i++)
    if (dists[i] % 2 == 0)
      ret += signs[i];
  return ret;
}

twopoly genalexander(code c) {
  if (c.n == 0) {
    twopoly zero;
    zero.n = 1;
    zero.m = 0;
    zero.terms = malloc(sizeof(poly));
    zero.terms->n = 1;
    zero.terms->m = 0;
    zero.terms->coeffs = malloc(sizeof(int));
    zero.terms->coeffs[0] = 0;
    return zero;
  }
  PD p = code2PD(c);
  twopoly ***mat = malloc(2 * c.n * sizeof(twopoly **));
  int i, j;
  for (i = 0; i < 2 * c.n; i++) {
    mat[i] = malloc(2 * c.n * sizeof(twopoly *));
    //for (j = 0; j < 2 * c.n; j++)
    //mat[i][j] = zero;
  }
  twopoly zero, mone, s, si, t, onest, ti, onesti;
  zero.n = mone.n = s.n = si.n = t.n = ti.n = 1;
  onest.n = onesti.n = 2;
  zero.m = mone.m = s.m = si.m = onest.m = 0;
  t.m = 1;
  ti.m = onesti.m = -1;
  zero.terms = malloc(sizeof(poly));
  zero.terms->n = 1;
  zero.terms->m = 0;
  zero.terms->coeffs = malloc(sizeof(int));
  zero.terms->coeffs[0] = 0;
  mone.terms = malloc(sizeof(poly)); // -1
  mone.terms->n = 1;
  mone.terms->m = 0;
  mone.terms->coeffs = malloc(sizeof(int));
  mone.terms->coeffs[0] = -1;
  s.terms = malloc(sizeof(poly));
  s.terms->n = 1;
  s.terms->m = 1;
  s.terms->coeffs = malloc(sizeof(int));
  s.terms->coeffs[0] = 1;
  si.terms = malloc(sizeof(poly)); // 1/s
  si.terms->n = 1;
  si.terms->m = -1;
  si.terms->coeffs = malloc(sizeof(int));
  si.terms->coeffs[0] = 1;
  t.terms = malloc(sizeof(poly));
  t.terms->n = 1;
  t.terms->m = 0;
  t.terms->coeffs = malloc(sizeof(int));
  t.terms->coeffs[0] = 1;
  onest.terms = malloc(2 * sizeof(poly)); // 1 - st
  onest.terms[0].n = onest.terms[1].n = 1;
  onest.terms[0].m = 0;
  onest.terms[0].coeffs = malloc(sizeof(int));
  onest.terms[0].coeffs[0] = 1;
  onest.terms[1].m = 1;
  onest.terms[1].coeffs = malloc(sizeof(int));
  onest.terms[1].coeffs[0] = -1;
  ti.terms = malloc(sizeof(poly)); // 1/t
  ti.terms->n = 1;
  ti.terms->m = 0;
  ti.terms->coeffs = malloc(sizeof(int));
  ti.terms->coeffs[0] = 1;
  onesti.terms = malloc(2 * sizeof(poly)); // 1 - 1/(st)
  onesti.terms[0].n = onesti.terms[1].n = 1;
  onesti.terms[0].m = -1;
  onesti.terms[0].coeffs = malloc(sizeof(int));
  onesti.terms[0].coeffs[0] = -1;
  onesti.terms[1].m = 0;
  onesti.terms[1].coeffs = malloc(sizeof(int));
  onesti.terms[1].coeffs[0] = 1;

  for (i = 0; i < 2 * c.n; i++)
    for (j = 0; j < 2 * c.n; j++)
      mat[i][j] = &zero;
  for (i = 0; i < 2 * c.n; i++) { // this is probably not optimal
    int n = c.code[i] / 6;
    if (c.code[i] % 3 == 2) { // positive crossing
      if (crossings[c.code[i]].p == O) { // over
	mat[i][p.crossings[n][3]] = &s;
	mat[i][p.crossings[n][1]] = &mone;
      } else {
	mat[i][p.crossings[n][3]] = &onest;
	mat[i][p.crossings[n][0]] = &t;
	mat[i][p.crossings[n][2]] = &mone;
      }
    } else {
      if (crossings[c.code[i]].p == O) {
	mat[i][p.crossings[n][1]] = &si;
	mat[i][p.crossings[n][3]] = &mone;
      } else {
	mat[i][p.crossings[n][1]] = &onesti;
	mat[i][p.crossings[n][0]] = &ti;
	mat[i][p.crossings[n][2]] = &mone;
      }
    }
  }
  twopoly ret = det(mat, 2 * c.n, &zero);
  freetwopoly(zero);
  freetwopoly(mone);
  freetwopoly(s);
  freetwopoly(t);
  freetwopoly(si);
  freetwopoly(ti);
  freetwopoly(onest);
  freetwopoly(onesti);
  for (i = 0; i < 2 * c.n; i++)
    free(mat[i]);
  free(mat);
  freePD(p);
  return ret;
}

twopoly det(twopoly ***mat, int n, twopoly *zero) {
  int used[n], permut[n], *zeros[n], i, j;
  for (i = 0; i < n; i++) {
    used[i] = 0;
    zeros[i] = malloc(n * sizeof(int));
    for (j = 0; j < n; j++)
      zeros[i][j] = mat[i][j] == zero;
  }
  twopoly ret;
  ret.n = 1;
  ret.m = 0;
  ret.terms = malloc(sizeof(poly));
  ret.terms->n = 1;
  ret.terms->m = 0;
  ret.terms->coeffs = malloc(sizeof(int));
  ret.terms->coeffs[0] = 0;
  detr(mat, n, &ret, permut, 0, used, zeros);
  for (i = 0; i < n; i++)
    free(zeros[i]);
  return ret;
}

void detr(twopoly ***mat, int n, twopoly *r, int *permut, int m, int *used,
	  int **zeros) { // recursive; does work for det
  int i;
  if (m == n) { // done creating this permutation
    int j, pn = 0;
    for (i = 0; i < n; i++) // find the permutation number
      for (j = 0; j < i; j++)
	if (permut[j] > permut[i])
	  pn++;
    twopoly term;
    term.n = 1;
    term.m = 0;
    term.terms = malloc(sizeof(poly));
    term.terms->n = 1;
    term.terms->m = 0;
    term.terms->coeffs = malloc(sizeof(int));
    term.terms->coeffs[0] = pn % 2 == 1 ? -1 : 1;
    for (i = 0; i < n; i++) {
      twopoly x = multtwopoly(term, *mat[i][permut[i]]);
      freetwopoly(term);
      term = x;
    }
    twopoly x = addtwopoly(*r, term);
    freetwopoly(term);
    freetwopoly(*r);
    *r = x;
  } else
    for (i = 0; i < n; i++)
      if (!used[i] && !zeros[m][i]) {
	used[i] = 1;
	permut[m] = i;
	detr(mat, n, r, permut, m + 1, used,  zeros);
	used[i] = 0;
      }
}

twopoly genalexhmirror(twopoly p) { // s -> s^-1; t -> t^-1
  twopoly r;
  r.n = p.n;
  r.m = -(p.n - 1 + p.m);
  r.terms = malloc(r.n * sizeof(poly));
  int i;
  for (i = 0; i < r.n; i++)
    r.terms[i] = jonesmirror(p.terms[p.n - 1 - i]);
  return r;
}

twopoly genalexvmirror(twopoly p) { // s -> t; t -> s
  twopoly r;
  r.n = 1;
  r.m = 0;
  r.terms = malloc(sizeof(poly));
  r.terms->n = 1;
  r.terms->m = 0;
  r.terms->coeffs = malloc(sizeof(int));
  r.terms->coeffs[0] = 0;

  int i;
  for (i = 0; i < p.n; i++) {
    twopoly a;
    a.n = p.terms[i].n;
    a.m = p.terms[i].m;
    a.terms = malloc(a.n * sizeof(poly));
    int j;
    for (j = 0; j < a.n; j++) {
      a.terms[j].n = 1;
      a.terms[j].m = i + p.m;
      a.terms[j].coeffs = malloc(sizeof(int));
      a.terms[j].coeffs[0] = p.terms[i].coeffs[j];
    }
    twopoly tmp = addtwopoly(r, a);
    freetwopoly(r);
    freetwopoly(a);
    r = tmp;
  }
  return r;
}

poly addpoly(poly a, poly b) {
  poly r;
  r.m = ((a.m < b.m) ? a.m : b.m);
  r.n = ((a.m + a.n > b.m + b.n) ? a.n + a.m - r.m : b.n + b.m - r.m);
  r.coeffs = calloc(r.n, sizeof(int));
  int i;
  for (i = 0; i < a.n; i++)
    r.coeffs[i + a.m - r.m] = a.coeffs[i];
  for (i = 0; i < b.n; i++)
    r.coeffs[i + b.m - r.m] += b.coeffs[i];
  return r;
}

twopoly addtwopoly(twopoly a, twopoly b) {
  twopoly r;
  r.m = ((a.m < b.m) ? a.m : b.m);
  r.n = ((a.m + a.n > b.m + b.n) ? a.n + a.m - r.m: b.n + b.m - r.m);
  r.terms = malloc(r.n * sizeof(poly));
  int i;
  for (i = 0; i < r.n; i++) {
    r.terms[i].n = 1;
    r.terms[i].m = 0;
    r.terms[i].coeffs = malloc(sizeof(int));
    r.terms[i].coeffs[0] = 0;
  }
  for (i = 0; i < a.n; i++) {
    //poly p = addpoly(r.terms[i + a.m - r.m], a.terms[i]);
    free(r.terms[i + a.m - r.m].coeffs);
    r.terms[i + a.m - r.m].n = a.terms[i].n;
    r.terms[i + a.m - r.m].m = a.terms[i].m;
    r.terms[i + a.m - r.m].coeffs = malloc(a.terms[i].n * sizeof(int));
    memcpy(r.terms[i + a.m - r.m].coeffs, a.terms[i].coeffs,
	   a.terms[i].n * sizeof(int));
  }
  for (i = 0; i < b.n; i++) {
    poly p = addpoly(r.terms[i + b.m - r.m], b.terms[i]);
    free(r.terms[i + b.m - r.m].coeffs);
    r.terms[i + b.m - r.m] = p;
  }
  return r;
}

poly multpoly(poly a, poly b) {
  poly r;
  r.m = a.m + b.m;
  r.n = a.n + b.n - 1;
  r.coeffs = calloc(r.n, sizeof(int));
  int i, j;
  for (i = 0; i < a.n; i++)
    if (a.coeffs[i] != 0)
      for (j = 0; j < b.n; j++)
	r.coeffs[i + j] += a.coeffs[i] * b.coeffs[j];
  return r;
}

twopoly multtwopoly(twopoly a, twopoly b) {
  twopoly r;
  r.m = a.m + b.m;
  r.n = a.n + b.n - 1;
  r.terms = malloc(r.n * sizeof(poly));
  poly zero;
  zero.m = 0;
  zero.n = 1;
  zero.coeffs = malloc(sizeof(int));
  zero.coeffs[0] = 0;
  int i, j;
  for (i = 0; i < r.n; i++) {
    r.terms[i].m = 0;
    r.terms[i].n = 1;
    r.terms[i].coeffs = malloc(sizeof(int));
    r.terms[i].coeffs[0] = 0;
  }
  for (i = 0; i < a.n; i++)
    if (comparepoly(a.terms[i], zero) != 0)
      for (j = 0; j < b.n; j++) {
	poly m = multpoly(a.terms[i], b.terms[j]);
	poly a = addpoly(r.terms[i + j], m);
	free(m.coeffs);
	free(r.terms[i + j].coeffs);
	r.terms[i + j] = a;
      }
  free(zero.coeffs);
  return r;
}

void freetwopoly(twopoly p) {
  int i;
  for (i = 0; i < p.n; i++)
    free(p.terms[i].coeffs);
  free(p.terms);
}

void initpascal(int n) {
  pascaltriangle = malloc(n * sizeof(int *));
  int i;
  for (i = 0; i < n; i++) {
    pascaltriangle[i] = malloc((i + 1) * sizeof(int));
    int j;
    for (j = 0; j <= i; j++) {
      if (j == 0 || j == i)
	pascaltriangle[i][j] = 1;
      else
	pascaltriangle[i][j] = pascaltriangle[i - 1][j - 1]
	  + pascaltriangle[i - 1][j];
    }
  }
}

int nCr(int n, int r) { // returns n!/((n-r)!r!); used by loopspoly()
  if (r < (n - r))
    r = n - r;
  int i;
  long long ret = 1;
  for (i = n; i > n - r; i--)
    ret *= i;
  for (i = 2; i <= r; i++)
    ret /= i;
  return (int) ret;
}

poly loopspoly(int n) { // (-A^2 - A^(-2))^n (assumes n >= 0)
  poly p;
  p.m = -2 * n;
  p.n = 4 * n + 1;
  p.coeffs = calloc(p.n, sizeof(int));
  int i;
  for (i = 0; i <= n; i++) {
    p.coeffs[4 * i] = pascaltriangle[n][i]; //nCr(n, i);
    if (n % 2 == 1)
      p.coeffs[4 * i] *= -1;
  }
  return p;
}

int comparepoly(poly a, poly b) { // returns 1 if a>b, -1 if a<b, 0 if a==b
  int i, j;
  for (j = 0; j < a.m - b.m && j < b.n; j++) {
    if (b.coeffs[j] > 0)
      return -1;
    if (b.coeffs[j] < 0)
      return 1;
  }
  for (i = 0; i < b.m - a.m && i < a.n; i++) {
    if (a.coeffs[i] < 0)
      return -1;
    if (a.coeffs[i] > 0)
      return 1;
  }
  for (; i < a.n && j < b.n; i++, j++) {
    if (a.coeffs[i] < b.coeffs[j])
      return -1;
    if (a.coeffs[i] > b.coeffs[j])
      return 1;
  }
  for (; i < a.n; i++) {
    if (a.coeffs[i] < 0)
      return -1;
    if (a.coeffs[i] > 0)
      return 1;
  }
  for (; j < b.n; j++) {
    if (b.coeffs[j] > 0)
      return -1;
    if (b.coeffs[j] < 0)
      return 1;
  }
  return 0;
}

int comparetwopoly(twopoly a, twopoly b) { // compares up to a factor of +/-s,t
  int ai, aj, bi, bj, am, bm, nega, negb, za = 0;
  for (ai = 0; ai < a.n; ai++)
    for (aj = 0; aj < a.terms[ai].n; aj++)
      if (a.terms[ai].coeffs[aj] != 0)
	goto ctp1;
 ctp1:
  if (ai == a.n)
    za = 1;
  else {
    bm = aj + a.terms[ai].m;
    nega = a.terms[ai].coeffs[aj] < 0;
  }
  for (bi = 0; bi < b.n; bi++)
    for (bj = 0; bj < b.terms[bi].n; bj++)
      if (b.terms[bi].coeffs[bj] != 0)
	goto ctp2;
 ctp2:
  if (za)
    if (bi == b.n)
      return 0;
    else
      return -1;
  if (bi == b.n)
    return 1;
  am = bj + b.terms[bi].m;
  negb = b.terms[bi].coeffs[bj] < 0;
  for (; ai < a.n && bi < b.n; ai++, bi++) { // TODO!!! handle a = -b
    //a.terms[ai].m += am;                   // !!! how to sort????
    //b.terms[bi].m += bm;                   // --> normalize so 1st term is >0
    poly ca = a.terms[ai], cb = b.terms[bi];
    ca.m += am;
    cb.m += bm;
    if (nega) {
      ca.coeffs = malloc(ca.n * sizeof(int));
      int i;
      for (i = 0; i < ca.n; i++)
	ca.coeffs[i] = -a.terms[ai].coeffs[i];
    }
    if (negb) {
      cb.coeffs = malloc(cb.n * sizeof(int));
      int i;
      for (i = 0; i < cb.n; i++)
	cb.coeffs[i] = -b.terms[bi].coeffs[i];
    }
    //int cmp = comparepoly(a.terms[ai], b.terms[bi]);
    //a.terms[ai].m -= am;
    //b.terms[bi].m -= bm;
    int cmp = comparepoly(ca, cb);
    if (nega)
      free(ca.coeffs);
    if (negb)
      free(cb.coeffs);
    if (cmp != 0)
      return cmp;
  }
  poly zero;
  zero.n = 1;
  zero.m = 0;
  zero.coeffs = malloc(sizeof(int));
  zero.coeffs[0] = 0;
  for (; ai < a.n; ai++) {
    int cmp = comparepoly(a.terms[ai], zero);
    if (cmp != 0) {
      free(zero.coeffs);
      return cmp;
    }
  }
  for (; bi < b.n; bi++) {
    int cmp = comparepoly(zero, b.terms[bi]);
    if (cmp != 0) {
      free(zero.coeffs);
      return cmp;
    }
  }
  free(zero.coeffs);
  return 0;
}

void printpoly(poly p) {
  cprintpoly(p, 'A');
}

void cprintpoly(poly p, char c){//currently outputs nothing for zero polynomial
  int i;
  for (i = 0; i < p.n; i++) {
    if (p.coeffs[i] == 0)
      continue;
    if (p.coeffs[i] == -1)
      printf("-");
    else if (p.coeffs[i] == 1)
      printf("+");
    else
      printf("%+d", p.coeffs[i]);
    if (i + p.m != 0) {
      printf("%c", c);
      if (i + p.m != 1)
	printf("^%d", i + p.m);
    } else if (p.coeffs[i] == 1 || p.coeffs[i] == -1)
      printf("1");
  }
}

char *sprintpoly(poly p) {
  return scprintpoly(p, 'A');
}

char *scprintpoly(poly p, char c) {
  char *s = malloc((30 * (p.n + 1)) * sizeof(char));
  int i, j = 0, printed = 0;
  for (i = 0; i < p.n; i++) {
    if (p.coeffs[i] == 0)
      continue;
    if (p.coeffs[i] == -1)
      j += sprintf(s + j, "- ");
    else if (p.coeffs[i] == 1) {
      if (printed)
	j += sprintf(s + j, "+ ");
    } else
      if (p.coeffs[i] > 0)
	if (printed)
	    j += sprintf(s + j, "+ %d", p.coeffs[i]);
	else
	  j += sprintf(s + j, "%d", p.coeffs[i]);
      else
	j += sprintf(s + j, "- %d", -p.coeffs[i]);
    if (i + p.m != 0) {
      j += sprintf(s + j, "%c", c);
      if (i + p.m != 1)
	j += sprintf(s + j, "<sup>%d</sup>", i + p.m);
    } else if (p.coeffs[i] == 1 || p.coeffs[i] == -1)
      j += sprintf(s + j, "1");
    j += sprintf(s + j, " ");
    printed = 1;
  }
  if (s[j - 1] == ' ')
    s[j - 1] = '\0';
  return s;
}

char *sprintjones(poly p) {
  char *s = malloc((35 * (p.n + 1)) * sizeof(char));
  int i, j = 0, printed = 0;
  for (i = p.n - 1; i >= 0; i--) {
    if (p.coeffs[i] == 0)
      continue;
    if (p.coeffs[i] == -1)
      j += sprintf(s + j, "- ");
    else if (p.coeffs[i] == 1) {
      if (printed)
	j += sprintf(s + j, "+ ");
    } else
      if (p.coeffs[i] > 0)
	if (printed)
	    j += sprintf(s + j, "+ %d", p.coeffs[i]);
	else
	  j += sprintf(s + j, "%d", p.coeffs[i]);
      else
	j += sprintf(s + j, "- %d", -p.coeffs[i]);
    if (i + p.m != 0) {
      j += sprintf(s + j, "q");
      if (i + p.m != 1)
	switch (abs(i + p.m) % 4) {
	case 0:
	  j += sprintf(s + j, "<sup>%d</sup>", -(i + p.m)/4);
	  break;
	case 2:
	  j += sprintf(s + j, "<sup>%d/2</sup>", -(i + p.m)/2);
	  break;
	default:
	  j += sprintf(s + j, "<sup>%d/4</sup>", -(i + p.m));
	}
    } else if (p.coeffs[i] == 1 || p.coeffs[i] == -1)
      j += sprintf(s + j, "1");
    j += sprintf(s + j, " ");
    printed = 1;
  }
  if (s[j - 1] == ' ')
    s[j - 1] = '\0';
  return s;
}

void printtwopoly(twopoly p) {
  int i, printed = 0;
  poly zero;
  zero.n = 1;
  zero.m = 0;
  zero.coeffs = malloc(sizeof(int));
  zero.coeffs[0] = 0;
  for (i = 0; i < p.n; i++) {
    if (comparepoly(p.terms[i], zero) == 0)
      continue;
    printed = 1;
    printf("+(");
    cprintpoly(p.terms[i], 's');
    printf(")");
    if (i + p.m != 0) {
      printf("t");
      if (i + p.m != 1)
	printf("^%d", i + p.m);
    }
  }
  if (!printed)
    printf("0");
  free(zero.coeffs);
}

char *sprinttwopoly(twopoly p) {
  int i, printed = 0, j = 0;
  for (i = 0; i < p.n; i++)
    j += 30 * (p.terms[i].n + 1) + 20;
  char *s = malloc(j * sizeof(char));
  j = 0;
  poly zero;
  zero.n = 1;
  zero.m = 0;
  zero.coeffs = malloc(sizeof(int));
  zero.coeffs[0] = 0;
  for (i = 0; i < p.n; i++) {
    if (comparepoly(p.terms[i], zero) == 0)
      continue;
    char *sp = scprintpoly(p.terms[i], 's');
    if (printed)
      j += sprintf(s + j, " + (%s)", sp);
    else
      j += sprintf(s + j, "(%s)", sp);
    free(sp);
    if (i + p.m != 0) {
      j += sprintf(s + j, " t");
      if (i + p.m != 1)
	j += sprintf(s + j, "<sup>%d</sup>", i + p.m);
    }
    printed = 1;
  }
  if (!printed)
    sprintf(s + j, "0");
  free(zero.coeffs);
  return s;
}

void printlink(link l) {
  int i;
  for (i = 0; i < l.n; i++) {
    int j;
    for (j = 0; j < l.codes[i].n; j++) {
      printf((crossings[l.codes[i].code[j]].p == U) ? "U" : "O");
      printf("%d", l.codes[i].code[j] / 6 + 1);
      if (crossings[l.codes[i].code[j]].s != 0)
	printf((crossings[l.codes[i].code[j]].s > 0) ? "+" : "-");
    }
    printf("/");
  }
  printf("\n");
}
