#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include "gauss.h"
#include "moves.h"
#include "pd.h"
#include "poly.h"
#include "draw.h"
#include "polytrees.h"
#include "writehtml.h"

int main(int argc, char *argv[]) {
  if (argc < 3) {
    printf("Usage: getknots <max crossings> <number of crossings for output> [0 for non-cyclic] [1 for reversible]\n");
    return 0;
  }
  int sz = atoi(argv[1]);
  int maxp = atoi(argv[2]);
  cyclic = (argc >= 4 ? atoi(argv[3]) : 1);
  reversible = (argc == 5 ? atoi(argv[4]) : 0);
  initcrossings(50 * maxp);
  initpascal(52 * maxp);
  knot n;
  code *codes = genallcodes(maxp, &n);
  //int *lowest = findlowest(codes, n, sz);
  //findlowest(codes, n, sz);
  //initlowest(FILENAME, 0);
  initlowest(sz);
  n <<= sz - maxp;

  FILE *file = fopen("html/index.html", "w");
  fprintf(file, "<!DOCTYPE html PUBLIC \"-//W3C/DTD XHTML 1.0 Strict //EN\""
	  " \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\n"
	  "<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\""
	  " lang=\"en\">\n"
	  "<head>\n"
	  "<link href=\"index.css\" rel=\"stylesheet\" type=\"text/css\" />\n"
	  "<title>Virtual Knot Table</title>\n"
	  "</head><body>\n"
	  "<div style=\"position: absolute; top: 0.3em; right: 0.5em;\">"
	  "Next: <a href=\"contents.html\">Contents</a></div>\n"
	  "<div style=\"position: absolute; top: 0.3em; right: 8em;\">"
	  "Up: <a href=\"../index.html#GreenJ\">Students</a></div>\n"
	  "<h1>A Table of Virtual Knots</h1>\n"
	  "<h4>Last updated August 10, 2004</h4>\n"
	  "<h3>Jeremy Green</h3>\n"
	  "<h5>(comments/corrections to "
	  "<a href=\"mailto:jeremy.green@utoronto.ca\">"
	  "jeremy.green@utoronto.ca</a>)</h5>\n"
	  "<p>This is a table of virtual knots. They were enumerated "
	  "using a computer program written by Jeremy Green under the "
	  "supervision of <a href=\"http://www.math.toronto.edu/~drorbn/\">"
	  "Dror Bar-Natan</a>. For more information about this project, see "
	  "the <a href=\"about.html\">about page</a> or the "
	  "<a href=\"results.html\">results</a> of the enumeration</a>. "
	  "There is also a <a href=\"contents.html\">table of contents</a> "
	  "for the associated pages.</p>");

  knot i, m = 0, j = 0;
  int nc = -1, nk = 1;
  kname names[n];
  polytree tjones, tjones3;
  twopolytree tgenalex;
  tjones.p.n = tjones3.p.n = tgenalex.p.n = -1;
  tjones.left = tjones.right = tjones3.left = tjones3.right = NULL;
  tgenalex.left = tgenalex.right = NULL;
  for (i = 0; i < n; i++) {
    names[i].nc = -1;
    names[i].nk = -1;
    code c = getcode(codes, sz, i);
    if (c.n != -1) {
      if (c.n <= maxp) {
	knot l = getlowest(i);
	if (l == -1) {
	  fprintf(stderr, "%d of %d\r", i, n);

	  knot f = getfirst(i, codes, n, sz);
	  if (f < i) {
	    names[i] = names[f];
	    free(c.code);
	    continue;
	  }

	  if (c.n > nc) {
	    nc = c.n;
	    nk = 1;
	    //fprintf(file, "<br />\n");
	    if (nc == 0 || nc > 3) {
	      if (nc == 0)
		fprintf(file, "<h2>Virtual knots with up to 3 crossings</h2>\n"
			"<table><tr>");
	      else
		fprintf(file, "</tr></table>\n"
			"<h2>Virtual knots with %d crossings</h2>\n"
			"<table>", nc);
	    }
	  }

	  if (nc > 3 && nk % 10 == 1) {
	    if (nk > 1)
	      fprintf(file, "</tr>");
	    fprintf(file, "<tr>");
	  }

	  //if (!reversible)
	  //printf("%c", f & 4 ? 'i' : '-');
	  //printf("%d ", f & 3);
	  //printcode(c);

	  names[i].nc = nc;
	  names[i].nk = nk;
	  poly p = jonespoly2(c);
	  if (tjones.p.n == -1) {
	    tjones.p = p;
	    tjones.codes = calloc(1, sizeof(knotlist));
	    tjones.codes->n = i;
	  } else
	    if (!addtotree(&tjones, p, i))
	      free(p.coeffs);
	  p = multijonespoly(c, 3);
	  if (tjones3.p.n == -1) {
	    tjones3.p = p;
	    tjones3.codes = calloc(1, sizeof(knotlist));
	    tjones3.codes->n = i;
	  } else
	    if (!addtotree(&tjones3, p, i))
	      free(p.coeffs);
	  twopoly tp = genalexander(c);
	  if (tgenalex.p.n == -1) {
	    tgenalex.p = tp;
	    tgenalex.codes = calloc(1, sizeof(knotlist));
	    tgenalex.codes->n = i;
	  } else
	    if (!addtotwotree(&tgenalex, tp, i))
	      freetwopoly(tp);

	  char name[64];
 	  snprintf(name, 63, "%d.%d", nc, nk);
/* 	  writeHTML(c, name, f); */
	  fprintf(file, "<td><a href=\"%s.html\"><div><a href=\"%s.html\">"
		  "<img src=\"images/%d.%d-s.png\" /></a><br />"
		  "<a href=\"%s.html\">%s</a>"
		  "</div></a></td>\n", name, name, nc, nk, name, name);

	  struct stat filestat;
	  snprintf(name, 63, "html/images/%d.%d.png", nc, nk);
	  if (stat(name, &filestat) != 0)
	    drawPNG(name, c, 250, 250);
	  snprintf(name, 63, "html/images/%d.%d-s.png", nc, nk);
	  if (stat(name, &filestat) != 0)
	    drawPNG(name, c, 64, 64);
	  nk++;

	  j++;
	}
      }
      free(c.code);
    }
  }
  fprintf(file, "</body></html>\n");
  fclose(file);

  fprintf(stderr, "                           \r");
  kname prevk = {-1, -1}, nextk;
  for (i = 0; i < n; i++) {
    if (cmpkname(prevk, names[i]) < 0) {
      code c = getcode(codes, sz, i);
      //if (c.n != -1) {
      //if (c.n <= maxp) {
      //knot l = getlowest(i);
      //if (l == -1) {
      fprintf(stderr, "%d of %d\r", i, n);
      
      int f = isfirst(i, codes, n, sz);
      if (f == -1) {
	free(c.code);
	continue;
      }
      
      knot a;
      for (a = i + 1; a < n; a++)
	if (cmpkname(names[i], names[a]) < 0) {
	  nextk = names[a];
	  break;
	}
      if (a == n) {
	nextk.nc = -1;
	nextk.nk = -1;
      }
	
      writeHTML(c, names[i], prevk, nextk, f, names, &tjones, &tjones3,
		&tgenalex);
      prevk = names[i];
	  //}
	  //}
      free(c.code);
      if (a == n)
	break;
      i = a - 1;
    }
  }

  printf("Number of knots: " KNOT_PRINTF "; output: " KNOT_PRINTF "\n", m, j);

  return 0;
}
